/* Generated automatically from kernel.h.in by configure. */
extern        void    init_kmem __P((char * file));
extern        int     klookup __P((int off, char * target, int siz));

#define KERNEL_IPSTAT "_ipstat"
#define KERNEL_IPFORWARDING "_ip_forwarding"
#define KERNEL_TCP_TTL "_tcp_ttl"
#define KERNEL_UDPSTAT "_udpstat"
#define KERNEL_IN_INTERFACES "_in_interfaces"
#define KERNEL_ICMPSTAT "_icmpstat"
#define KERNEL_IFNET "_ifnet"
#define KERNEL_TCPSTAT "_tcpstat"
#define KERNEL_TCB "_tcb"
#define KERNEL_ARPTAB_SIZE "_arptab_size"
#define KERNEL_ARPTAB "_arptab"
#define KERNEL_IN_IFADDR "_in_ifaddr"
#define KERNEL_BOOTTIME "_boottime"
#define KERNEL_PROC "_proc"
#define KERNEL_NPROC "_nproc"
#define KERNEL_DMMIN "unknown"
#define KERNEL_DMMAX "unknown"
#define KERNEL_NSWAP "unknown"
#define KERNEL_USRPTMAP "unknown"
#define KERNEL_USRPT "unknown"
#define KERNEL_USERSIZE "unknown"
