/* acconfig.h

   Descriptive text for the C preprocessor macros that
   the locally-defined Autoconf macros can define.

   The entries are in sort -df order: alphabetical, case insensitive,
   ignoring punctuation (such as underscores).  Although this order
   can split up related entries, it makes it easier to check whether
   a given entry is in the file.

   Leave the following blank line there!!  Autoheader needs it.  */



/* Define if struct ortentry is defined */
#undef  HAVE_STRUCT_ORTENTRY

/* Define the following if sys_errlist exists */
#undef  HAVE_SYS_ERRLIST

/* Define the following if sys_nerr is declared */
#undef  HAVE_SYS_ERRLIST_DECL

/* Define the following if sys_nerr exists */
#undef  HAVE_SYS_NERR

/* Define the following if sys_nerr is declared */
#undef  HAVE_SYS_NERR_DECL

/* Define if _KMEMUSER needed, e.g. by UnixWare */ 
#undef  _KMEMUSER

/* On systems that lack a typedef for socklen_t, define it as unsigned int.
   Should really be defined as uint32_t, but we define it
   to be an unsigned int, as it is needed early in the compile process,
   sometimes before some implementations define uint32_t. */
#undef  socklen_t       /* <sys/socket.h> */

/* Define if STREAMS ioctls should be used on socket descriptors, instead of SIOCxxx ioctls */
#undef  SYS_SOCKET_IOCTLS_USE_STREAMS


/* Leave that blank line there!!  Autoheader needs it.
   If you're adding to this file, keep in mind:
   The entries are in sort -df order: alphabetical, case insensitive,
   ignoring punctuation (such as underscores).  */

@BOTTOM@

/* Define if struct sockaddr{} has sa_len member */
#undef  STRUCT_SOCKADDR_HAS_SA_LEN

